#define FORMAT_C TRUE
/* ========================================================================
 * Name - format.c
 *
 * Description - 
 *
 * $Id: format.c,v 1.2 1992/09/15 22:37:01 penneyj Exp $
 *
 * $Log: format.c,v $
 * Revision 1.2  1992/09/15  22:37:01  penneyj
 * misc facelift, preparing for distribution on alt.sources
 *
 * Revision 1.1  1992/09/15  22:18:47  penneyj
 * Initial revision
 *
 *========================================================================
 */

#include "general.h"
#include "config.h"

#include "list.h"
#include "main.h"
#include "utils.h"

#include "format.h" /* this file's interface */

#if defined(FLG_PROTOTYPE)
static BoolType uncompressFile(const char *filein, const char *fileout)
#else
static BoolType uncompressFile(filein, fileout)
     const char *filein, *fileout;
#endif
{
  char *buffer;
  const char *fstr = "zcat <%s >%s";
  BoolType result;
     
  buffer = MyMalloc(strlen(fstr) + strlen(filein) + strlen(fileout));
  sprintf(buffer, fstr, filein, fileout);
  result = doit(buffer, NULL);
  FREE(buffer);
  return result;
}

#if defined(FLG_PROTOTYPE)
static BoolType unpackFile(const char *filein, const char *fileout)
#else
static BoolType unpackFile(filein, fileout)
     const char *filein, *fileout;
#endif
{
  char *buffer;
  const char *fstr = "pcat %s >%s";
  BoolType result;
     
  buffer = MyMalloc(strlen(fstr) + strlen(filein) + strlen(fileout));
  sprintf(buffer, fstr, filein, fileout);
  result = doit(buffer, NULL);
  FREE(buffer);
  return result;
}

/* compose an nroff command... */
#if defined(FLG_PROTOTYPE)
static const char *nroffcommand(const char *filein, const char *fileout)
#else
static const char *nroffcommand(filein, fileout)
     const char *filein, *fileout;
#endif
{
  BoolType  neqn, refer, tbl, vgrind, col;
  int extrabufsize;
  char *buffer, *extrab;
  FILE *infile;
  const char *fstr =
	"cat %s %s | neqn | refer | tbl | vgrind | nroff | col > %s";

  extrabufsize = (strlen(fstr) + strlen(manmacros) + strlen(filein) +
      strlen(fileout));
  buffer = MyMalloc(extrabufsize);     
  extrab = MyMalloc(extrabufsize);
     
  infile = fopen(filein, "r");
  fgets(buffer, 10, infile);
  fclose(infile);

  if (NULL == strchr(buffer, '\n'))
    *buffer = '\0' ;
  if (strncmp(buffer, "'\\\" ", 4) == 0) {
    col = neqn = (NULL != strchr(buffer, 'e'));
    col |= tbl = (NULL != strchr(buffer, 't'));
    refer =      (NULL != strchr(buffer, 'r'));
    vgrind =     (NULL != strchr(buffer, 'v'));
    }
  else
    neqn = refer = tbl = vgrind = col = FALSE;

  /* Col may be necessary anyway. sigh. */
  col = TRUE; /* Testing this value */
  tbl = TRUE;
  neqn = TRUE;
  sprintf(buffer, "cat %s %s", manmacros, filein);

#if !defined(FLG_ULTRIX_UNIX) /* ultrix has no eqn, refer, or vgrind */
  if (neqn) {
    sprintf(extrab, "%s | neqn", buffer);
    strcpy(buffer, extrab);
    }
  if (refer) {
    sprintf(extrab, "%s | refer", buffer);
    strcpy(buffer, extrab);
    }
#endif
  if (tbl) {
    sprintf(extrab, "%s | tbl", buffer);
    strcpy(buffer, extrab);
    }
#if !defined(FLG_ULTRIX_UNIX)
  if (vgrind) {
    sprintf(extrab, "%s | vgrind", buffer);
    strcpy(buffer, extrab);
    }
#endif
     
  sprintf(extrab, "%s | nroff", buffer);
  strcpy(buffer, extrab);
     
  if (col) {
    sprintf(extrab, "%s | col", buffer);
    strcpy(buffer, extrab);
    }
  sprintf(extrab, "%s > %s", buffer, fileout);
  strcpy(buffer, extrab);     

  return buffer;
}


#if defined(FLG_PROTOTYPE)
char *makeCatName(const char oldName[], const char catDir[],
    const char baseName[])
#else
char *makeCatName(oldName, catDir, baseName)
const char oldName[];
const char catDir[];
const char baseName[];
#endif
{
  char *newName;
  char *suffix;

  /* printf("oldName=%s catDir=%s baseName=%s\n", oldName, catDir, baseName); */
  newName = MyMalloc(strlen(catDir) + strlen(oldName) + EXTRASPACE);
  strcpy(newName, catDir);
  suffix = strrchr(newName, '.');

  /* if last path component contains a .Z or a .z, remove it */
  if (suffix != NULL && strchr(suffix, '/') == NULL)
    *suffix = '\0';
  sprintf(&newName[strlen(newName)], "%s%s", DIRSEP, baseName);
  if (debug_f)
    fprintf(stderr, "catName for %s is %s\n", oldName, newName);
  return newName;
}

/* Checks each file for .so directives */
#if defined(FLG_PROTOTYPE)
static BoolType checkforso(ListType item)
#else
static BoolType checkforso(item) 
ListType item;
#endif
{
  FILE *nr_file;   /* For all nroff files */
  long filelength;
  char linebuf[81], *line;
  char *oldName = (*item)->key;
  char *newName;

  line = &linebuf[0];
  nr_file = fopen(oldName, "r");
  if (nr_file == NULL) {
    warning("Couldn't open %s", oldName);
    perror(oldName);
    return FALSE;
    }

  fseek(nr_file, 0, 2);
  filelength = ftell(nr_file);
  /* 12 is the shortest possible - I think */
  if (filelength < 12) {
    warning("Too short file: %s", oldName);
    fclose(nr_file);
    return FALSE;
    }
  fseek(nr_file, 0, 0);

  /* Read first line of each file */
  fgets(line, 80, nr_file);
  if (strncmp(line, ".so ", 4) != 0) {
    fclose(nr_file);
    return TRUE;
    }

  if (debug_f)
    fprintf(stderr, ".so found in file %s:\n\"%s\"", oldName, line);
  line += 4;
  if (strchr(line, ' ') != NULL) 
    *(strchr(line, ' ')) = '\0';
  if (strchr(line, '\n') != NULL) 
    *(strchr(line, '\n')) = '\0';
  newName = MyMalloc(strlen((*item)->dirname) + strlen(line) + EXTRASPACE);
  sprintf(newName, "%s%s%s", (*item)->dirname, DIRSEP, line);
  (*item)->key = newName;
  if (debug_f)
    fprintf(stderr, "Replacing with %s\n", newName);

  if (strncmp(line, "man", 3) != 0 || line[4] != '/') {
    warning("man[1-8lnop]/... expected in .so line \"%s\"", line - 4);
    (*item)->catname = NULL;
    fclose(nr_file);
    return TRUE;
    }

  if ((*item)->catname != NULL) {
    char *baseName = strrchr((*item)->key, '/');

    *baseName = '\0'; /* null-terminate the cat-dir */
    baseName ++;
    /* pick a catfile in same directory, different name */
    (*item)->catname = makeCatName(baseName, (*item)->key, &line[3]);
    }
  fclose(nr_file);
  return TRUE;
}

/*
 *  Checks for duplicate entries, and for already
 *  nroff'ed entries.
 */
#if defined(FLG_PROTOTYPE)
static void notneeded(void)
#else
static void notneeded() 
#endif
{
  ListType item, already;
  char *this;
  long filelength, checksum;
  struct stat statbuf;
  time_t last_ch;
  FILE *testfile;

  already = newlist();
  LIST_DO (matchingFiles) {
    if ((*item)->kind != NROFF_PAGE && (*item)->kind != ZCAT_NROFF_PAGE)
      continue;

    this = (*item)->key;    /* Should not be null. */
    if (stat(this, &statbuf) != 0) {
      fprintf(stderr, "Stat failed on ");
      perror(this);
      KILLTHIS(item);
      continue;
      }
    last_ch = statbuf.st_mtime; /* Last change to the man file */
    if (every_f) {
      ADDTOLIST(already, this); /* No more checking on duplication */
      continue;
      }

    /* Smaller than any possible man file? */
    filelength = statbuf.st_size;
    if (filelength < 12) {
      warning("%s: file too small", this);
      KILLTHIS(item);
      continue;
      }
    testfile = fopen(this, "r");
    if (testfile == NULL) {
      warning("Couldn't open file %s for reading", this);
      perror(this);
      KILLTHIS(item);
      continue;
      }
    fseek(testfile, filelength / 2, 0); /* CCL */
    fread((char *)&checksum, sizeof(checksum), 1, testfile); /* CCL */
    fclose(testfile);
    if (inlist23(already, (char *)filelength, (char *)checksum) && !every_f) {
      /* Probably duplicate entry */
      if (verbos_f)
	warning("Uninteresting file %s", this);
      KILLTHIS(item);
      continue;
      }

    add3list(already, this, (char *)filelength, (char *)checksum, NO_PAGE);
    this = (*item)->catname;
    if (this != NULL)
      if (stat(this, &statbuf) == 0)
	/* Do we have a newer 'cat' file ? */		    
        if (statbuf.st_mtime > last_ch && !reform_f && statbuf.st_size > 0) {
	  if (debug_f)
	    printf("discarding %s in favor of newer cat-file %s\n",
		(*item)->key, this);
	  (*item)->kind = CAT_PAGE;
	  (*item)->key = this;
	  }
    if (only1_f) {
      (*item)->next = NULL;
      return;
      }
    }
  cleannulls(matchingFiles);
}

/* cover to invoke nroff on one file */
#if defined(FLG_PROTOTYPE)
static BoolType grindFile(ListType item)
#else
static BoolType grindFile(item)
ListType item;
#endif
{
  FILE *infile, *outfile;
  const char *todo;
  char *inName;
  char *outName;

  if (!checkforso(item))
    return FALSE;
  if (!checkforso(item))
    return FALSE;
  /* how many times should we do this before we get disgusted? */

  inName = (*item)->key;
  if ((*item)->catname == NULL)
    (*item)->catname = (char *)tmpfilename();
  outName = (*item)->catname;
  outfile = fopen(outName, "w");
  if (outfile == NULL) {
    (*item)->catname = outName = (char *)tmpfilename();
    outfile = fopen(outName, "w");
    if (outfile == NULL) {
      warning("Fatal error - couldn't open %s", outName);
      die(FILERROR);
      }
    }

  notyet = outName;
  infile = fopen(inName, "r");
  if (infile == NULL) {
    warning("Couldn't open %s", inName);
    fprintf(outfile,"%s%s%s", "Couldn't open ", inName,
	" -- somebody muffed it\n");
    fclose(outfile);
    return FALSE;
    }	  
  fclose(outfile);
  fclose(infile);
  todo = nroffcommand(inName, outName);
  if (debug_f) 
    fprintf(stderr, "nroffing %s into %s\n", inName, outName);
  if (!doit(todo, "Reformatting page."))
    return FALSE;
  return TRUE;
}

/* all the work -- invoke programs on all found files... */
#if defined(FLG_PROTOTYPE)
void formatfiles(void)
#else
void formatfiles()
#endif
{
  ListType item;

  notneeded();
  LIST_DO (matchingFiles) {
    switch ((*item)->kind) {
      case CAT_PAGE:
	ADDTOLIST(showfiles, (*item)->catname);
	break;

      case PACKED_CAT_PAGE:
	(*item)->catname = (char *)tmpfilename();
	if (!unpackFile((*item)->key, (*item)->catname))
	  break;
	ADDTOLIST(showfiles, (*item)->catname);
        break;
	
      case COMPRESSED_CAT_PAGE:
	(*item)->catname = (char *)tmpfilename();
	if (!uncompressFile((*item)->key, (*item)->catname))
	  break;
	ADDTOLIST(showfiles, (*item)->catname);
        break;

      case NROFF_PAGE:
        if (!grindFile(item))
	  break;
	ADDTOLIST(showfiles, (*item)->catname);
	break;

      case ZCAT_NROFF_PAGE: {
	char *newName = (char *)tmpfilename();

	if (!uncompressFile((*item)->key, newName))
	  break;
	(*item)->key = newName; /* let system grind on uncompressed file */
        if (!grindFile(item))
	  break;
	ADDTOLIST(showfiles, (*item)->catname);
	break;
	}

      default:
	warning("Fatal error - unknown type of file, %d",
	    (void *)(*item)->kind);
	die(INTERNERROR);
      } /* switch */
    } /* LIST_DO */
}
